################################################################################
## jigebren's Re-Volt import / exporter
################################################################################
"""
Import and export Re-Volt FIN, FOB, HUL, LIT, NCP, PRM/M, RIM, TAZ, VIS and W 
file formats. Can also dump several of them to a TXT file. Also import 
parameters.txt file, and BMP textures (with black pixels transparency support). 
Add a Re-Volt section to the Data panel.
"""

bl_info = {
    "name": "Re-Volt", 
    "author": "jigebren", 
    "version": (15, 2, 27), 
    "blender": (2, 73, 0), 
    "location": "File > Import-Export", 
    "description": "Import and export FIN, FOB, HUL, LIT, NCP, PRM/M, RIM, TAZ, VIS and W file formats.", 
    "warning": "", 
    "wiki_url": "", 
    "tracker_url": "http://z3.invisionfree.com/Our_ReVolt_Pub/index.php?showtopic=1500", 
    "category": "Import-Export",
    }

######## import ########

from . import common
from . import panel

if "bpy" in locals():
    import imp
    
    imp.reload(common)
    imp.reload(panel)
    
    if "io_all" in locals(): imp.reload(io_all)
    if "io_car" in locals(): imp.reload(io_car)
    if "io_fin" in locals(): imp.reload(io_fin)
    if "io_fob" in locals(): imp.reload(io_fob)
    if "io_hul" in locals(): imp.reload(io_hul)
    if "io_img" in locals(): imp.reload(io_img)
    if "io_lit" in locals(): imp.reload(io_lit)
    if "io_ncp" in locals(): imp.reload(io_ncp)
    if "io_prm" in locals(): imp.reload(io_prm)
    if "io_rim" in locals(): imp.reload(io_rim)
    if "io_rtu" in locals(): imp.reload(io_rtu)
    if "io_taz" in locals(): imp.reload(io_taz)
    if "io_vis" in locals(): imp.reload(io_vis)
    if "io_w"   in locals(): imp.reload(io_w)

import bpy
import os
import os.path
import time
from .common import *

######## Scene RV properties ########

class RV_SettingsScene(bpy.types.PropertyGroup):
    
    # global settings
    
    scale = bpy.props.FloatProperty(
            name = "Scale", 
            description="Scaling down factor applied to the imported data (scaling up when exporting)", 
            default = 100, 
            min = 1, max = 10000, 
            )
    prm_zoom_enable = bpy.props.BoolProperty(
            name = "10x factor for cars, .hul, .prm", 
            description="Make cars, instances, etc. 10x times bigger for easier editing", 
            default = True, 
            )
    update_ncp_with_w = bpy.props.BoolProperty(
            name = "Also export .ncp", 
            description="Export the NCP collision file at the same time", 
            default = False, 
            )
    color_key_to_alpha = bpy.props.BoolProperty(
            name = "ColorKey transparency", 
            description="Update textures so that black pixels are made transparent similarly to Re-Volt (slower)", 
            default = True, 
            )
    import_car_hull = bpy.props.BoolProperty(
            name = "Import Car Collisions", 
            description="Also import the .hul file and the wheels collision spheres", 
            default = False, 
            )
    import_bigcubes = bpy.props.BoolProperty(
            name = "Import BigCubes", 
            description="Create BigCubes groups and Bounding Spheres", 
            default = IMPORT_WORLD_BIGCUBES, 
            )
    dump_mode = bpy.props.BoolProperty(
            name = "Dump Mode", 
            description="Dump the data to a readable text file (data are not imported)", 
            default = False, 
            )
    ncp_non_planar_threshold = bpy.props.FloatProperty(
            name = "Dist", 
            description="Acceptable face 'thickness' before it is considered non-planar (in Re-Volt unit)", 
            default = 1, 
            min = 0.01, soft_min = 1, soft_max = 10, max = 100, 
            step = 100, precision = 2
            )
    last_exported_filepath = bpy.props.StringProperty(subtype='FILE_PATH')
    
    # expand boxes in the RV panel
    
    expand_car = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=False)
    expand_fin = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=False)
    expand_fob = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=False)
    expand_hul = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=False)
    expand_img = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=False)
    expand_lit = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=False)
    expand_ncp = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=False)
    expand_prm = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=False)
    expand_rim = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=False)
    expand_rtu = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=False)
    expand_taz = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=False)
    expand_vis = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=False)
    expand_w   = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=False)
    expand_vertex_paint = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=True)
    expand_texanim      = bpy.props.BoolProperty(name="", description="Show / Hide panel", default=True)
    
    # edit mode tab
    
    edit_mode_format = bpy.props.EnumProperty(
            name="Face edit mode", 
            description="Select the edit mode", 
            items=(
                    ("prm", "PRM / World", "(.prm/.m; .w)"), 
                    ("ncp", "NCP", "Collision (.ncp)"), 
                    ), 
            default="prm", 
            )
    
    # face rendering flags
    
    face_flag = bpy.props.EnumProperty(
            name="Rendering Flag", 
            description="Select a flag for Batch Editing", 
            items=(
                    ("8192", "Invisible", ""), 
                    (   "2", "Double-sided", ""), 
                    (   "4", "Translucent", ""), 
                    ( "128", "Mirroring", ""), 
                    ( "256", "Add Blending", ""), 
                    ( "512", "TexAnim", ""), 
                    ("1024", "No EnvMap (.prm)", ""), 
                    ("2048", "EnvMap (.w)", ""), 
                    ("4096", "SoftFx", ""), 
                    ), 
            default="2", 
            )
    
    # TexAnim
    
    texanim_frame0 = bpy.props.StringProperty(  # we're using a String to be able to store a list (via repr() and eval())
            name="TexAnim slots frame 0", 
            description="First frame definition for each TexAnim slots", 
            # default=repr([[0, 1, (0,) * 8]] * MAX_TEXANIMS), 
            default=repr([()] * MAX_TEXANIMS), 
            )
    texanim_slot = bpy.props.IntProperty(
            name="TexAnim Slot", 
            description="The TexAnim Slot to work with", 
            default=0, 
            min=0, max=MAX_TEXANIMS - 1,
            update=update_texanim_slot, 
            )
    texanim_time = bpy.props.FloatProperty(
            name = "Time", 
            description="Default frame duration for this TexAnim slot (in seconds)", 
            default = 1, 
            min = 0, max = 10, 
            update=update_texanim_time, 
            # set=set_texanim_time, 
            # get=get_texanim_time, 
            )
    
    # vertex color mode (paint / select)
    
    vertex_color_mode = bpy.props.EnumProperty(
            name="Mode", 
            description="Select the edit mode", 
            items=(
                    ("paint", "Paint", "Paint color on selected vertices / edges / faces"), 
                    ("select", "Select", "Select vertices / edges / faces by color"), 
                    ), 
            default="paint", 
            )
    
    # face ncp flags
    
    ncp_flag = bpy.props.EnumProperty(
            name="NCP Flag", 
            description="Select a NCP flag for Batch Editing", 
            items=(
                    ## ("2", "NCP_TWOSIDED", ""),  # this property is actually not supported by re-volt
                    ( "4", "Object Only", ""), 
                    ( "8", "Camera Only", ""), 
                    ("16", "Non Planar", ""), 
                    ("32", "No Skid", ""), 
                    ("64", "Oil", ""), 
                    ), 
            default="4", 
            )
    
    # face ncp material
    
    ncp_material = bpy.props.EnumProperty(
            name="Material", 
            description="Select a material for Batch Editing", 
            items=(
                    ("-1", "NONE", ""), 
                    ( "0", "DEFAULT", ""), 
                    ( "1", "MARBLE", ""), 
                    ( "2", "STONE", ""), 
                    ( "3", "WOOD", ""), 
                    ( "4", "SAND", ""), 
                    ( "5", "PLASTIC", ""), 
                    ( "6", "CARPETTILE", ""), 
                    ( "7", "CARPETSHAG", ""), 
                    ( "8", "BOUNDARY", ""), 
                    ( "9", "GLASS", ""), 
                    ("10", "ICE1", ""), 
                    ("11", "METAL", ""), 
                    ("12", "GRASS", ""), 
                    ("13", "BUMPMETAL", ""), 
                    ("14", "PEBBLES", ""), 
                    ("15", "GRAVEL", ""), 
                    ("16", "CONVEYOR1", ""), 
                    ("17", "CONVEYOR2", ""), 
                    ("18", "DIRT1", ""), 
                    ("19", "DIRT2", ""), 
                    ("20", "DIRT3", ""), 
                    ("21", "ICE2", ""), 
                    ("22", "ICE3", ""), 
                    ("23", "WOOD2", ""), 
                    ("24", "CONVEYOR_MARKET1", ""), 
                    ("25", "CONVEYOR_MARKET2", ""), 
                    ("26", "PAVING", ""), 
                    ), 
            default="0", 
            )
    
    # active face ncp material
    
    active_ncp_material_last = bpy.props.StringProperty(default="0", options={'HIDDEN'})
    active_ncp_material = bpy.props.EnumProperty(
            name="Material", 
            description="Set the active face material", 
            items=(
                    ("-1", "NONE"               , "", -1), 
                    ( "0", "DEFAULT"            , "",  0), 
                    ( "1", "MARBLE"             , "",  1), 
                    ( "2", "STONE"              , "",  2), 
                    ( "3", "WOOD"               , "",  3), 
                    ( "4", "SAND"               , "",  4), 
                    ( "5", "PLASTIC"            , "",  5), 
                    ( "6", "CARPETTILE"         , "",  6), 
                    ( "7", "CARPETSHAG"         , "",  7), 
                    ( "8", "BOUNDARY"           , "",  8), 
                    ( "9", "GLASS"              , "",  9), 
                    ("10", "ICE1"               , "", 10), 
                    ("11", "METAL"              , "", 11), 
                    ("12", "GRASS"              , "", 12), 
                    ("13", "BUMPMETAL"          , "", 13), 
                    ("14", "PEBBLES"            , "", 14), 
                    ("15", "GRAVEL"             , "", 15), 
                    ("16", "CONVEYOR1"          , "", 16), 
                    ("17", "CONVEYOR2"          , "", 17), 
                    ("18", "DIRT1"              , "", 18), 
                    ("19", "DIRT2"              , "", 19), 
                    ("20", "DIRT3"              , "", 20), 
                    ("21", "ICE2"               , "", 21), 
                    ("22", "ICE3"               , "", 22), 
                    ("23", "WOOD2"              , "", 23), 
                    ("24", "CONVEYOR_MARKET1"   , "", 24), 
                    ("25", "CONVEYOR_MARKET2"   , "", 25), 
                    ("26", "PAVING"             , "", 26), 
                    ), 
            default="0", 
            #update=update_active_ncp_material, 
            set=set_active_ncp_material, 
            get=get_active_ncp_material, 
            )
    
    # cast shadow type
    
    shadow_type = bpy.props.EnumProperty(
            name="Shadow Type", 
            description="Choose the type of shadow to cast", 
            items=(
                    ("current", "Use current setup", "Don't modify the current Lamp location and Shadow or Bake settings"), 
                    ("sharp", "Sharp", "Create a sharp (ie. aliased) shadow"), 
                    ("smooth", "Smooth", "Shadow is anti-aliased"), 
                    ("soft", "Soft", "Shadow is more blurred than 'Smooth' preset"), 
                    ("soft-far", "Soft (far)", "Car is mover further from the ground, giving an even more blurred shadow"), 
                    ("ao", "Ambient occlusion", "Use ambient occlusion instead of casting shadow from a light"), 
                    ("ao-far", "Ambient occlusion (far)", "Car is mover further from the ground, giving an even more blurred shadow"), 
                    ), 
            default="current", 
            update=update_shadow_type, 
            )

######## Object RV properties ########

"""NOTE: we're not using Mesh properties anymore, as they're not copied to the 
mesh we'll work on when calling to_mesh().
"""

class RV_SettingsObject(bpy.types.PropertyGroup):
    
    # fin
    
    is_fin = bpy.props.BoolProperty(
            # name="Object is an Instance", 
            description="Object is an Instance", 
            # description="Only Instance objects are exported to the .fin file (and automatically rejected from World and World NCP file)", 
            )
    fin_col = bpy.props.FloatVectorProperty(
            name="Model Col", 
            description="Model Color", 
            default=(0.0, 0.0, 0.0), 
            min=0.0, soft_min=0.0, max=1.0, soft_max=1.0, 
            subtype='COLOR', 
            )
    fin_envcol = bpy.props.FloatVectorProperty(
            name="EnvCol", 
            description="Env Color", 
            default=(0.5, 0.5, 0.5), 
            min=0.0, soft_min=0.0, max=1.0, soft_max=1.0, 
            subtype='COLOR', 
            )
    fin_priority = bpy.props.BoolProperty(
            name="High Priority", 
            description="High priority instances are not disabled even when instances are turned off", 
            )
    fin_flag_env = bpy.props.BoolProperty(
            name="Env Col", 
            description="Enable Env Color", 
            )
    fin_flag_hide = bpy.props.BoolProperty(
            name="Hide", 
            description="Hide", 
            )
    fin_flag_no_mirror = bpy.props.BoolProperty(
            name="No Mirror", 
            description="No Mirror", 
            )
    fin_flag_no_lights = bpy.props.BoolProperty(
            name="No Lights", 
            description="No File Lights", 
            )
    fin_flag_no_object_coll = bpy.props.BoolProperty(
            name="No Object Coll", 
            description="No Object Collision", 
            )
    fin_flag_no_camera_coll = bpy.props.BoolProperty(
            name="No Camera Coll", 
            description="No Camera Collision", 
            )
    fin_lod_bias = bpy.props.IntProperty(
            name="LoD Bias", 
            description="LoD Bias", 
            default=1024, 
            min=64, soft_min=64, max=8192, soft_max=8192, 
            )
    
    # fob
    
    is_fob = bpy.props.BoolProperty(
            # name="Object is a Re-Volt object", 
            description="Object is a Re-Volt object", 
            # description="Only Re-Volt objects are exported to the .fob file", 
            )
    """fob_id = bpy.props.IntProperty(name="ID", description="Object ID", 
            default=0, soft_min=0, soft_max=OBJECT_TYPE_MAX, 
            update=update_fob_id)"""
    fob_id = bpy.props.EnumProperty(
            name="ID", 
            description="Re-Volt object type (WARNING: not all objects are allowed in a FOB file, this property should rather be edited in MAKEITGOOD)", 
            items=list_to_items(FOB_NAME), 
            default="-1", 
            update=update_fob_id)
    fob_flag0 = bpy.props.IntProperty(name="flag 0", description="Flag 0 (WARNING: this property depends on the object ID and should rather be edited in MAKEITGOOD)")
    fob_flag1 = bpy.props.IntProperty(name="flag 1", description="Flag 1 (WARNING: this property depends on the object ID and should rather be edited in MAKEITGOOD)")
    fob_flag2 = bpy.props.IntProperty(name="flag 2", description="Flag 2 (WARNING: this property depends on the object ID and should rather be edited in MAKEITGOOD)")
    fob_flag3 = bpy.props.IntProperty(name="flag 3", description="Flag 3 (WARNING: this property depends on the object ID and should rather be edited in MAKEITGOOD)")
    
    # hul
    
    is_hul_convex = bpy.props.BoolProperty(
            # name="Convex Hull", 
            description="Object is a Convex Hull", 
            update=update_is_hul_convex, 
            )
    is_hul_sphere = bpy.props.BoolProperty(
            # name="HUL collision sphere", 
            description="Object is a HUL collision sphere", 
            update=update_is_hul_sphere, 
            )
    
    # lit
    
    is_lit = bpy.props.BoolProperty(
            # name="Object is a Re-Volt Light", 
            description="Object is a Re-Volt Light", 
            # description="Only lights with this property are be exported to the .lit file", 
            update=update_is_lit, 
            )
    lit_type = bpy.props.EnumProperty(
            name="Type", 
            description="Re-Volt light type (this property should rather be edited in MAKEITGOOD)", 
            items=list_to_items(LIT_TYPE), 
            default=str(LIGHT_OMNI), 
            update=update_lit_type, 
            )
    lit_flag = bpy.props.IntProperty(
            name="Flags", description="Bit-fields: 1 (World) | 2 (Objects) | 4 (1) | 8 (0) | 16 (Flicker)", 
            default=LIGHT_FILE | LIGHT_FIXED | LIGHT_MOVING, soft_min=LIGHT_FIXED, soft_max=LIGHT_FLICKER*2-1, 
            )
    lit_speed = bpy.props.IntProperty(
            name="Flicker Speed", description="Flicker Speed", 
            default=1, soft_min=1, soft_max=255, 
            )
    
    # rim
    
    is_rim = bpy.props.BoolProperty(
            description="Object is a Mirror Plane", 
            update=update_is_rim, 
            )
    
    # taz
    
    is_taz = bpy.props.BoolProperty(
            description="Object is a Track Zone", 
            update=update_is_taz, 
            )
    taz_id = bpy.props.IntProperty(
            name="ID", description="TrackZone ID", 
            default=0, min=0, soft_min=0, max=1024, soft_max=256, 
            update=update_taz_id, 
            )
    
    # vis
    
    is_vis = bpy.props.BoolProperty(
            description="Object is a Visibox", 
            update=update_is_vis, 
            )
    vis_type = bpy.props.EnumProperty(
            name="Type", 
            description="Select the VisiBox type", 
            items=(
                    ('VISIBOX_CAMERA', "Camera", "When the camera is inside this box, parts of the world can be hidden using Cube boxes"), 
                    ('VISIBOX_CUBE', "Cube", "Everything inside this box is hidden when the camera is inside a Camera box using the same ID"), 
                    ), 
            default="VISIBOX_CUBE", 
            update=update_vis_type, 
            )
    vis_id = bpy.props.IntProperty(
            name="ID", 
            description="Camera and Cube boxes are linked together by sharing the same ID", 
            default=0, min=0, soft_min=0, max=VISIBOX_MAX_ID - 1, soft_max=VISIBOX_MAX_ID - 1, 
            update=update_vis_id, 
            )
    
    # w
    

################################################################################
## user interface
################################################################################

# formats

FORMAT_UNK   = -1
FORMAT_BMP   =  0
FORMAT_CAR   =  1
FORMAT_FIN   =  2
FORMAT_FOB   =  3
FORMAT_HUL   =  4
FORMAT_LIT   =  5
FORMAT_NCP   =  6
FORMAT_NCP_W =  7
FORMAT_PRM   =  8
FORMAT_RIM   =  9
FORMAT_RTU   = 10
FORMAT_TAZ   = 11
FORMAT_VIS   = 12
FORMAT_W     = 13

file_formats = {
    FORMAT_UNK   : "Unknown Format", 
    FORMAT_BMP   : "BMP", 
    FORMAT_CAR   : "Parameters.txt", 
    FORMAT_FIN   : "FIN", 
    FORMAT_FOB   : "FOB", 
    FORMAT_HUL   : "HUL", 
    FORMAT_LIT   : "LIT", 
    FORMAT_NCP   : "NCP (Object)", 
    FORMAT_NCP_W : "NCP (World)", 
    FORMAT_PRM   : "PRM/M", 
    FORMAT_RIM   : "RIM", 
    FORMAT_RTU   : "RTU", 
    FORMAT_TAZ   : "TAZ", 
    FORMAT_VIS   : "VIS", 
    FORMAT_W     : "W", 
}

def get_file_format(filepath):
    """Return presumed file_format according to filepath."""
    filename, ext = os.path.splitext(os.path.basename(filepath.lower()))
    
    if ext.startswith('.bm'):
        return FORMAT_BMP
    elif filename.startswith('parameters') and ext == '.txt':
        return FORMAT_CAR
    elif ext == '.fin':
        return FORMAT_FIN
    elif ext == '.fob':
        return FORMAT_FOB
    elif ext == '.hul':
        return FORMAT_HUL
    elif ext == '.lit':
        return FORMAT_LIT
    elif ext == '.ncp':
        # NCP World Mode if there's a .w file
        if os.path.isfile(os.path.splitext(filepath)[0] + ".w"):
            return FORMAT_NCP_W
        else:
            return FORMAT_NCP
    elif ext == '.prm' or ext == '.m':
        return FORMAT_PRM
    elif ext == '.rim':
        return FORMAT_RIM
    elif ext == '.rtu':
        return FORMAT_RTU
    elif ext == '.taz':
        return FORMAT_TAZ
    elif ext == '.vis':
        return FORMAT_VIS
    elif ext == '.w':
        return FORMAT_W
    return FORMAT_UNK

"""
format = {
        "car": "Parameters.txt", 
        "fin": "*.fin", 
        "hul": "*.hul", 
        "bmp": "*.bmp", 
        "ncp": "*.ncp", 
        "prm": "*.prm;*.m", 
        "rim": "*.rim", 
        "rtu": "*.rtu", 
        "taz": "*.taz", 
        "vis": "*.vis", 
        "w"  : "*.w", 
        }
def update_format(self, context):
    print(self)
    print(self.filter_glob)
    self.filter_glob = format[self.format]
    print(self.filter_glob)
    # And now update space (file select window), if possible.
    space = bpy.context.space_data
    # XXX Can't use direct op comparison, these are not the same objects!
    if (space.type != 'FILE_BROWSER' or space.operator.bl_rna.identifier != self.bl_rna.identifier):
        print("<<< BAD >>>")
        return
    space.params.filter_glob = self.filter_glob
    print(dir(space.params))
    self.filepath = space.params.filename
    # XXX Seems to be necessary, else not all changes above take effect...
    bpy.ops.file.refresh()
"""

class ImportRV(bpy.types.Operator):
    bl_idname = "import_scene.revolt"  # was using import.revolt but couldn't call this op (probably because import is a reserved word)
    bl_label = 'Import Re-Volt'
    bl_description = 'Import from several Re-Volt file formats'
    
    '''
    filename_ext = ".prm"
    filter_glob = bpy.props.StringProperty(default="*.prm", options={'HIDDEN'})
    '''
    filepath = bpy.props.StringProperty(subtype='FILE_PATH')
    
    # List of operator properties, the attributes will be assigned
    # to the class instance from the operator settings before calling.
    '''filepath = bpy.props.StringProperty(name="File Path", 
        description="Filepath used for importing the revolt file", 
        maxlen= 1024, default= "", subtype='FILE_PATH')'''
    
    def execute(self, context):
        scn = context.scene
        props = context.scene.revolt
        
        starttime = time.time()
        inform("\n--- jigebren's Re-Volt importer {20%02d/%02d/%02d} ---\n" % bl_info['version'])
        
        # call the import function corresponding to the file extension
        format = get_file_format(self.filepath)
        
        if format == FORMAT_UNK:
            inform("This file format is not supported!", True)
        elif format == FORMAT_BMP:
            from . import io_img
            io_img.import_file(self.filepath)
            inform('\nDone! Image imported in %.2f seconds.\n' % (time.time() - starttime))
        else:
            context.window.cursor_set('WAIT')
            # turn off undo for better performance
            use_global_undo = bpy.context.user_preferences.edit.use_global_undo
            bpy.context.user_preferences.edit.use_global_undo = False
            
            # ensure object mode
            if bpy.ops.object.mode_set.poll():
                bpy.ops.object.mode_set(mode='OBJECT')
            
            # only imported objects selected after import
            bpy.ops.object.select_all(action='DESELECT')
            
            # update global vars
            common.scale = props.scale
            
            # set zoom scale for formats that use it
            if (format in {FORMAT_CAR, FORMAT_HUL, FORMAT_NCP, FORMAT_PRM} and
                    props.prm_zoom_enable):
                common.scale /= PRM_ZOOM_FACTOR
            
            #TODO try something like this to see ReVolt panel by default:
            #row.operator("wm.properties_context_change", text="", icon='BUTS').context = 'RENDER_LAYER'
            # Also check PROPERTIES_HT_header to see if we can add a real ReVolt panel
            
            if format == FORMAT_CAR:
                from . import io_car
                io_car.import_file(self.filepath, scn, props.dump_mode)
            elif format == FORMAT_FIN:
                from . import io_fin
                io_fin.import_file(self.filepath, scn, props.dump_mode)
            elif format == FORMAT_FOB:
                from . import io_fob
                io_fob.import_file(self.filepath, scn, props.dump_mode)
            elif format == FORMAT_HUL:
                from . import io_hul
                io_hul.import_file(self.filepath, scn, props.dump_mode)
            elif format == FORMAT_LIT:
                from . import io_lit
                io_lit.import_file(self.filepath, scn, props.dump_mode)
            elif format == FORMAT_NCP:
                from . import io_ncp
                io_ncp.import_file(self.filepath, scn, props.dump_mode)
            elif format == FORMAT_NCP_W:
                from . import io_ncp
                io_ncp.import_file(self.filepath, scn, props.dump_mode, world_mode=True)
            elif format == FORMAT_PRM:
                from . import io_prm
                io_prm.import_file(self.filepath, scn, props.dump_mode)
            elif format == FORMAT_RIM:
                from . import io_rim
                io_rim.import_file(self.filepath, scn, props.dump_mode)
            elif format == FORMAT_RTU:
                from . import io_rtu
                io_rtu.import_file(self.filepath, scn, props.dump_mode)
            elif format == FORMAT_TAZ:
                from . import io_taz
                io_taz.import_file(self.filepath, scn, props.dump_mode)
            elif format == FORMAT_VIS:
                from . import io_vis
                io_vis.import_file(self.filepath, scn, props.dump_mode)
            elif format == FORMAT_W:
                from . import io_w
                io_w.import_file(self.filepath, scn, props.dump_mode)
            
            # make Re-Volt Panel easier to find
            set_properties_context('DATA')
            
            # increase clip_end
            increase_clip_end(1000000 / common.scale)
            
            # move view to imported objects
            view_selected()  # user friendly
            
            # turn undo back on
            bpy.context.user_preferences.edit.use_global_undo = use_global_undo
            
            context.window.cursor_set('DEFAULT')
            inform('\nDone! Data imported in %.2f seconds.\n' % (time.time() - starttime))
        return {'FINISHED'}
    
    def invoke(self, context, event):
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}
        
    def draw(self, context):
        props = context.scene.revolt
        layout = self.layout
        space = context.space_data
        '''
        # TMP DEBUG, don't know whether it could actually happen or not...
        if (space.type != 'FILE_BROWSER' or space.operator.bl_rna.identifier != self.bl_rna.identifier):
            inform("<<< BAD >>>", True)
        '''
        # use the file selected in the file browser
        format = get_file_format(os.path.join(space.params.directory, space.params.filename))
        
        # only needed options according to selected file
        layout.label("Format: %s" % file_formats[format])
        if not format in {'FORMAT_UNK'}:
            if not format in {'FORMAT_BMP'}:
                layout.prop(props, 'scale')
                if format in {FORMAT_CAR, FORMAT_HUL, FORMAT_NCP, FORMAT_PRM}:
                    layout.prop(props, 'prm_zoom_enable')
            if format in {FORMAT_BMP, FORMAT_CAR, FORMAT_PRM, FORMAT_RIM, FORMAT_RTU, FORMAT_W}:
                layout.prop(props, 'color_key_to_alpha')
            if format in {FORMAT_CAR}:
                layout.prop(props, 'import_car_hull')
            if format in {FORMAT_W}:
                layout.prop(props, 'import_bigcubes')
            layout.prop(props, 'dump_mode')


class ExportRV(bpy.types.Operator):
    bl_idname = "export_scene.revolt"
    bl_label = 'Export Re-Volt'
    bl_description = 'Export to several Re-Volt file formats'
    
    filepath = bpy.props.StringProperty(subtype='FILE_PATH')
    
    def execute(self, context):
        scn = context.scene
        props = context.scene.revolt
        
        start_time = time.time()
        inform("\n--- jigebren's Re-Volt exporter {20%02d/%02d/%02d} ---\n" % bl_info['version'])
        
        # call the export function corresponding to the file extension
        format = get_file_format(self.filepath)
        
        if format == FORMAT_UNK:
            inform("This file format is not supported!", True)
        if format in {FORMAT_BMP, FORMAT_CAR}:
            #from . import io_img
            #from . import io_car
            inform("Exporting to %s is not supported." % file_formats[format])
        else:
            context.window.cursor_set('WAIT')
            # turn off undo for better performance
            use_global_undo = bpy.context.user_preferences.edit.use_global_undo
            bpy.context.user_preferences.edit.use_global_undo = False
            
            if bpy.ops.object.mode_set.poll():
                bpy.ops.object.mode_set(mode='OBJECT')
            
            # save filepath
            props.last_exported_filepath = self.filepath
            
            # update global vars
            common.scale = props.scale
            
            # set zoom scale for formats that use it
            if (format in {FORMAT_CAR, FORMAT_HUL, FORMAT_NCP, FORMAT_PRM} and
                    props.prm_zoom_enable):
                common.scale /= PRM_ZOOM_FACTOR
            
            if format == FORMAT_FIN:
                from . import io_fin
                io_fin.export_file(self.filepath, context)
            elif format == FORMAT_FOB:
                from . import io_fob
                io_fob.export_file(self.filepath, context)
            elif format == FORMAT_HUL:
                from . import io_hul
                io_hul.export_file(self.filepath, context)
            elif format == FORMAT_LIT:
                from . import io_lit
                io_lit.export_file(self.filepath, context)
            elif format == FORMAT_NCP:
                from . import io_ncp
                io_ncp.export_file(self.filepath, context)
            elif format == FORMAT_NCP_W:
                from . import io_ncp
                io_ncp.export_file(self.filepath, context, world_mode=True)
            elif format == FORMAT_PRM:
                from . import io_prm
                io_prm.export_file(self.filepath, context)
            elif format == FORMAT_RIM:
                from . import io_rim
                io_rim.export_file(self.filepath, context)
            elif format == FORMAT_RTU:
                #from . import io_rtu
                inform("Exporting to %s is not supported." % file_formats[format])
            elif format == FORMAT_TAZ:
                from . import io_taz
                io_taz.export_file(self.filepath, context)
            elif format == FORMAT_VIS:
                from . import io_vis
                io_vis.export_file(self.filepath, context)
            elif format == FORMAT_W:
                from . import io_w
                io_w.export_file(self.filepath, context)
                if props.update_ncp_with_w:
                    from . import io_ncp
                    io_ncp.export_file(os.path.splitext(self.filepath)[0] + '.ncp', context)
            
            # turn undo back on
            bpy.context.user_preferences.edit.use_global_undo = use_global_undo
            
            context.window.cursor_set('DEFAULT')
            inform('\nDone! Data exported in %.2f seconds.\n' % (time.time() - start_time))
        #TODO self.report({'INFO'}, "bla bla")
        return {'FINISHED'}
    
    def invoke(self, context, event):
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}
    
    def draw(self, context):
        props = context.scene.revolt
        layout = self.layout
        space = context.space_data
        # use the file selected in the file browser
        format = get_file_format(os.path.join(space.params.directory, space.params.filename))
        
        layout.label("Format: %s" % file_formats[format])
        layout.prop(props, 'scale')
        if format in {FORMAT_CAR, FORMAT_HUL, FORMAT_NCP, FORMAT_PRM}:
            layout.prop(props, 'prm_zoom_enable')
        if format == FORMAT_W:
            layout.prop(props, 'update_ncp_with_w')

################################################################################
## register
################################################################################

""" Template to create a sub menu in File...
class INFO_MT_file_ImportFT(bpy.types.Menu):
    bl_idname = "INFO_MT_file_ImportFT"
    bl_label = "FT import"

    def draw(self, context):
        layout = self.layout
        layout.operator_context = 'INVOKE_REGION_WIN'

        layout.operator(ImportFT.bl_idname, text="Fun Tracks (.*)")


def menu_func_import(self, context):
    self.layout.menu("INFO_MT_file_ImportFT", icon="PLUGIN")

def menu_func_export(self, context):
    self.layout.operator(ExportFT.bl_idname, text="Fun Tracks (.*)")

def register():
    bpy.utils.register_module(__name__)
    
    #bpy.utils.register_class(RV_SettingsScene)
    #bpy.utils.register_class(RV_SettingsObject)
    bpy.types.Scene.ft = bpy.props.PointerProperty(type=FT_SettingsScene)
    bpy.types.Object.ft = bpy.props.PointerProperty(type=FT_SettingsObject)
    bpy.types.Mesh.ft = bpy.props.PointerProperty(type=FT_SettingsMesh)
    
    bpy.types.INFO_MT_file.append(menu_func_import)
    bpy.types.INFO_MT_file_export.append(menu_func_export)
"""

def menu_func_import(self, context):
    self.layout.operator(ImportRV.bl_idname, text="Re-Volt (.fin; .fob; .hul; .lit; .ncp; .prm/.m; .rim; .taz; .vis; .w)")

def menu_func_export(self, context):
    self.layout.operator(ExportRV.bl_idname, text="Re-Volt (.fin; .fob; .hul; .lit; .ncp; .prm/.m; .rim; .taz; .vis; .w)")

def register():
    bpy.utils.register_module(__name__)
    
    #bpy.utils.register_class(RV_SettingsScene)
    #bpy.utils.register_class(RV_SettingsObject)
    bpy.types.Scene.revolt = bpy.props.PointerProperty(type=RV_SettingsScene)
    bpy.types.Object.revolt = bpy.props.PointerProperty(type=RV_SettingsObject)
    
    bpy.types.INFO_MT_file_import.prepend(menu_func_import) # DEV HACK: using prepend instead of append!
    bpy.types.INFO_MT_file_export.prepend(menu_func_export)

def unregister():
    bpy.utils.unregister_module(__name__)
    
    del bpy.types.Scene.revolt
    del bpy.types.Object.revolt

    bpy.types.INFO_MT_file_import.remove(menu_func_import)
    bpy.types.INFO_MT_file_export.remove(menu_func_export)

if __name__ == "__main__":
    register()

################################################################################
